;;; -*- Lisp -*-
;;;
;;; connection from Lisp to DPS window server and vice versa.
;;; This code (and the associated nps.c -- compiled cc nps.c -c)
;;;  provide window manipulations and mouse events,
;;;  and some functions to mimic Next's windows.

(in-package :cmn :use '(:lisp :loop))

#|
This code implements a simple tie into the Display Postscript
window system on the Next.  A new window is created by:

create-window x0 y0 x1 y1 &key title menu (close-button t) (vertical-scroll t) (horizontal-scroll t) (resize t)

|#

(defconstant screen-height 830)		;these apply to the Next
(defconstant screen-width 1120)



;;; Ties into NeXT specific constants and structs are via c procedures.
;;; Relevant header files are /usr/include/dspclient/*.h

#|

This code does not fit into the Next user interface very cleanly,
but it's not for lack of trying!  Here is a summary of the blind alleys 
I have encountered:

First there's no clean way to use the appkit from lisp --
if you follow the examples like Evaluator.cl on the lisp examples/objc
directory, you get only one such run per lisp run, and you can't
pass back and forth between lisp and the "application".  The "application"
takes over, and lisp is a subjob for the duration. If you
try to make a window and use the current NXApp (whatever it
may be -- Emacs or the Terminal), you get segmentation violations.
The only clean way I can see is to set up mach ports between the
lisp listener and a separate application running a postscript
viewer and sending mouse/key events back to lisp, but this 
pathway appears to me to be slow, and requires that I learn
a huge amount of NeXT-specific foolishness (the appkit, obj-c etc),
and I'm too old for that.

If you use raw Display PostScript code, as here, you can't easily get
appkit events or key events.  As far as I can tell, the application 
kit interferes at a low level with the event stream generated by the 
window server, and there is no documented way to get into that loop without
being a full-fledged application.  (This means, for example,
that I can't cause the title-bar to become gray upon deactivation
because I never see the deactivation event). If a key down event
occurs, it is passed to the lisp image, but is not immediately
available to read-char, listen, or read-char-no-hang -- apparently
the character stream is buffered, and only a newline flushes it.
So, our entire previewer has to be purely mouse-driven, and cannot
implement an object editor involving lisp code (unless we use Morse
code or break into the window loop using undocumented library code...)

Since I'm left to my own devices here, I've just tried to get as close
as I could to the Next style, but a few little differences are
still floating around.  (/usr/lib/NextStep/windowPackage.ps)
|#

#+Excl (progn

	 (load "~/cl/nps.o" :system-libraries '("sys_s" "dspops"))

	 (ff:defforeign 'dps-display-file :entry-point "_DisplayPSFile" :arguments '(string) :return-type :integer)
	 (ff:defforeign 'dps-create-context :entry-point "_CreatePSContext" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-kill-context :entry-point "_KillPSContext" :arguments nil :return-type :void)
	 (ff:defforeign 'dps-send :entry-point "_SendPS" :arguments '(string) :return-type :integer)
	 (ff:defforeign 'dps-flush :entry-point "_FlushPS" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-get-event-no-matter-what :entry-point "_GetPSEventNoMatterWhat" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-get-event :entry-point "_GetPSEvent" :arguments '(float) :return-type :integer)
	 
	 (ff:defforeign 'dps-mouse-down :entry-point "_MouseDown" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-mouse-up :entry-point "_MouseUp" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-mouse-down-mask :entry-point "_MouseDownMask" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-mouse-up-mask :entry-point "_MouseUpMask" :arguments nil :return-type :integer)
	 
	 (ff:defforeign 'dps-event-type :entry-point "_EventType" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-event-window :entry-point "_EventWindow" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-event-location-x :entry-point "_EventLocationX" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-event-location-y :entry-point "_EventLocationY" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-event-mouse-eventnum :entry-point "_EventDataMouseEventNum" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Flags :entry-point "_EventFlags" :arguments nil :return-type :integer)

	 (ff:defforeign 'dps-Event-Mouse-Click :entry-point "_EventDataMouseClick" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Key-Repeat :entry-point "_EventDataKeyRepeat" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Key-charSet :entry-point "_EventDataKeycharSet" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Key-charCode :entry-point "_EventDataKeycharCode" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Key-Data :entry-point "_EventDataKeyData" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Tracking-eventNum :entry-point "_EventDataTrackingeventNum" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Tracking-Num :entry-point "_EventDataTrackingNum" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Tracking-User-Data :entry-point "_EventDataTrackingUserData" :arguments nil :return-type :integer)
	 (ff:defforeign 'dps-Event-Compound-Subtype :entry-point "_EventDataCompoundSubtype" :arguments nil :return-type :integer)
	 ;; we could have used Franz's def-cstruct (or whatever it's called) for most of these
	 )



;;; ----------- icons
;;;
;;; some of the icons are very hard to duplicate, so I've stolen the bitmaps...

(defun close-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%15 15 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFF0~%EAAAAA90~%E1AAA490~%E46A9190~%E91A4690~%EA451A90~%EA906A90~%EA906A90~%~
                EA451A90~%E91A4690~%E46A9190~%E1AAA490~%EAAAAA90~%D5555550~%00000000 >} image grestore~%" x y)))

(defun half-close-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%15 15 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFF2~%EAAAAA92~%E1AAA492~%E46A9192~%E92A8692~%EAAAAA92~%EAAAAA92~%EAAAAA92~%~
                EAAAAA92~%E92A8692~%E46A9192~%E1AAA492~%EAAAAA92~%D5555552~%00000002 >} image grestore~%" x y)))

(defun right-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFC~%EAAAAAA4~%EAAAAAA4~%E9AAAAA4~%E81AAAA4~%E801AAA4~%E8001AA4~%E80001A4~%~
                E8001AA4~%E801AAA4~%E81AAAA4~%E9AAAAA4~%EAAAAAA4~%EAAAAAA4~%D5555554~%00000000 >} image grestore~%" x y)))

(defun inverse-right-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFF9~%FFFFFFF9~%FEFFFFF9~%FD6FFFF9~%FD56FFF9~%FD556FF9~%FD5556F9~%FD556FF9~%~
                FD56FFF9~%FD6FFFF9~%FEFFFFF9~%FFFFFFF9~%FFFFFFF9~%EAAAAAA9~%55555555~%55555555 >} image grestore~%" x y)))

(defun up-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFC~%EAAAAAA4~%EAAAAAA4~%EAA9AAA4~%EAA8AAA4~%EAA46AA4~%EAA02AA4~%EA901AA4~%~
                EA800AA4~%EA4006A4~%EA0002A4~%E90001A4~%EAAAAAA4~%EAAAAAA4~%D5555554~%00000000 >} image grestore~%" x y)))

(defun inverse-up-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFF9~%FFFFFFF9~%FFFEFFF9~%FFFDFFF9~%FFF9BFF9~%FFF57FF9~%FFE56FF9~%FFD55FF9~%~
                FF955BF9~%FF5557F9~%FE5556F9~%FFFFFFF9~%FFFFFFF9~%EAAAAAA9~%55555555~%55555555 >} image grestore~%" x y)))

(defun blank-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%15 15 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFC~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%~
                EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%D5555554 >} image grestore~%" x y)))

(defun checked-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%15 15 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFF~%FFFFFFF2~%EAAAAA92~%EAAAAB92~%EAAAAC92~%EAAAB192~%EB2AC692~%EB2B1A92~%~
                EB1C6A92~%EB71AA92~%EBC6AA92~%EB1AAA92~%E86AAA92~%EAAAAA92~%D5555552 >} image grestore~%" x y)))

(defun down-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFC~%EAAAAAA4~%EAAAAAA4~%E90001A4~%EA0002A4~%EA4006A4~%EA800AA4~%EA901AA4~%EAA02AA4~%~
                EAA46AA4~%EAA8AAA4~%EAA9AAA4~%EAAAAAA4~%EAAAAAA4~%D5555554~%00000000 >} image grestore~%" x y)))

(defun inverse-down-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFF9~%FFFFFFF9~%FE5556F9~%FF5557F9~%FF955BF9~%FFD55FF9~%FFE56FF9~%FFF57FF9~%~
                FFF9BFF9~%FFFDFFF9~%FFFEFFF9~%FFFFFFF9~%FFFFFFF9~%EAAAAAA9~%55555555~%55555555 >} image grestore~%" x y)))

(defun left-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFC~%EAAAAAA4~%EAAAAAA4~%EAAAA9A4~%EAAA90A4~%EAA900A4~%EA9000A4~%E90000A4~%EA9000A4~%~
                EAA900A4~%EAAA90A4~%EAAAA9A4~%EAAAAAA4~%EAAAAAA4~%D5555554~%00000000 >} image grestore~%" x y)))

(defun inverse-left-arrow-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFF9~%FFFFFFF9~%FFFFFEF9~%FFFFE5F9~%FFFE55F9~%FFE555F9~%FE5555F9~%FFE555F9~%~
                FFFE55F9~%FFFFE5F9~%FFFFFEF9~%FFFFFFF9~%FFFFFFF9~%EAAAAAA9~%55555555~%55555555 >} image grestore~%" x y)))

(defun dimple-button-image (x y)
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%16 16 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%FFFFFFFC~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EA902AA4~%EA455AA4~%EA16AAA4~%EA1AFAA4~%EA1BFAA4~%~
                EA9BEAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%EAAAAAA4~%D5555554~%00000000 >} image grestore~%" x y)))

(defun dimple-image (x y)		;i.e just the dimple without the button border
  (dps-send (format nil "gsave ~D ~D translate~%16.000 15.125 scale~%12 12 2~%[16 0 0 15.125 neg 0 15.125]~%~
            {<~%AAAAAA~%AAAAAA~%A902AA~%A455AA~%A16AAA~%A1AFAA~%A1BFAA~%~
                A9BEAA~%AAAAAA~%AAAAAA~%AAAAAA~%AAAAAA >} image grestore~%" x y)))



;;; ---------- window paraphenalia

(defstruct window x0 y0 x1 y1 scale lx0 ly0 title active id 
	   close-button title-bar unsaved-changes rx0 ry0 rx1 ry1 
	   title-bar-rect
	   close-button-rect
	   resize-left-button-rect
	   resize-right-button-rect
	   scroll-up-button-rect
	   scroll-down-button-rect
	   scroll-left-button-rect
	   scroll-right-button-rect
	   vertical-scroll-rect
	   horizontal-scroll-rect
	   menu-rect
	   menu-rects
	   user-rect)

(defun unprotect-window (win)
  (dps-send (format nil "~D ~D ~D ~D rectviewclip~%" 
		    (window-x0 win) (window-y0 win) (window-x1 win) (window-y1 win))))

(defun protect-window (win)
  (dps-send (format nil "~D ~D ~D ~D rectviewclip~%"
		    (window-rx0 win) (window-ry0 win) (window-rx1 win) (window-ry1 win))))

(defun unclip-window (win)
  (declare (ignore win))
  (dps-send (format nil "initviewclip~%")))

(defun clear-window (win)
  (dps-send (format nil "1 setgray ~D ~D ~D ~D rectfill 0 setgray~%"
		    (window-rx0 win) (window-ry0 win) 
		    (- (window-rx1 win) (window-rx0 win))
		    (if (not (window-menu-rect win))
			(- (window-ry1 win) (window-ry0 win))
		      (- (rect-y0 (window-menu-rect win)) (window-ry0 win))))))

(defun draw-title-bar (window active)
  (let ((wx0 (window-x0 window))
	(wx1 (window-x1 window))
	(wy1 (window-y1 window))
	(title (window-title window)))
    (setf (window-active window) active)
    (dps-send (format nil "~1,3F setgray 1 ~D ~D ~D rectfill~%~1,3F setgray 1 ~D ~D ~D rectfill~%"
		      (if active .667 1)
		      (- wy1 3) (- wx1 2) 2
		      (if active .333 0)
		      (- wy1 23) (- wx1 2) 20))
    (dps-send (format nil "~1,3F setgray~%~
                           ~D ~D ~D ~D rectfill gsave~%~
                           ~D ~D moveto~%~
                           ~D setgray~%~
                           /~A findfont ~D scalefont setfont~%~
                           (~A) show grestore~%~
                           0 setgray~%~
                           ~D ~D moveto ~D 0 rlineto stroke~%"
		      (if active 0 .667)
		      (+ wx0 2) (- wy1 21) (- wx1 wx0 4) 19
		      (- (floor (+ wx0 wx1) 2) (* 3 (length title)))
		      (- wy1 16)
		      (if active 1 0)
		      "Helvetica-Bold" 14
		      title
		      wx0 (- wy1 22) wx1))
    (if (window-unsaved-changes window)
	(half-close-button-image (- wx1 20) (- wy1 20))
      (close-button-image (- wx1 20) (- wy1 20)))))

(defun draw-vertical-scroll-bar-slot (window)
  (let ((wx0 (window-x0 window))
	(wy0 (window-y0 window))
	(wy1 (window-y1 window))
	(y 0))
    (setf (window-vertical-scroll-rect window) (make-rect :x0 (+ wx0 1) :y0 (+ wy0 8) :x1 (+ wx0 18) :y1 (- wy1 24)))
    (dps-send (format nil ".5 setgray ~D ~D ~D ~D rectfill~%~
                           .667 setgray ~D ~D moveto ~D 0 rlineto 0 ~D rlineto ~D 0 rlineto stroke~%~
                           0 setgray ~D ~D moveto 0 ~D rlineto stroke~%"
		      (+ wx0 1) (+ wy0 8) 17 (setf y (- wy1 wy0 8 24))
		      (+ wx0 1) (- wy1 23) 17 (- (- y 1)) -17
		      (+ wx0 19) (+ wy0 9) y))))

(defun draw-horizontal-scroll-bar-slot (window)
  (let ((wx0 (window-x0 window))
	(wx1 (window-x1 window))
	(wy0 (window-y0 window))
	(x 0))
    (setf (window-horizontal-scroll-rect window) (make-rect :x0 (+ wx0 20) :y0 (+ wy0 8) :x1 (- wx1 1) :y1 (+ wy0 25)))
    (dps-send (format nil ".5 setgray ~D ~D ~D ~D rectfill~%~
                           .667 setgray ~D ~D moveto ~D 0 rlineto 0 ~D rlineto ~D 0 rlineto stroke~%~
                           0 setgray ~D ~D moveto ~D 0 rlineto stroke~%"
		      (+ wx0 20) (+ wy0 8) (setf x (- wx1 wx0 21)) 18
		      (+ wx0 21) (+ wy0 26) (- x 2) -17 (- (- x 2))
		      (+ wx0 20) (+ wy0 27) (+ x 2)))))

(defun draw-vertical-scroll-buttons (window)
  (let ((wx0 (window-x0 window))
	(wy0 (window-y0 window)))
    (up-arrow-button-image (+ wx0 2) (+ wy0 10 17))
    (down-arrow-button-image (+ wx0 2) (+ wy0 10))
    (setf (window-scroll-up-button-rect window) (make-rect :x0 (+ wx0 2) :y0 (+ wy0 10 17) :x1 (+ wx0 2 16) :y1 (+ wy0 10 17 16)))
    (setf (window-scroll-down-button-rect window) (make-rect :x0 (+ wx0 2) :y0 (+ wy0 10) :x1 (+ wx0 2 16) :y1 (+ wy0 10 16)))))

(defun draw-horizontal-scroll-buttons (window)
  (let ((wx0 (window-x0 window))
	(wy0 (window-y0 window)))
    (left-arrow-button-image (+ wx0 22) (+ wy0 10))
    (right-arrow-button-image (+ wx0 22 17) (+ wy0 10))
    (setf (window-scroll-left-button-rect window) (make-rect :x0 (+ wx0 22) :y0 (+ wy0 10) :x1 (+ wx0 22 16) :y1 (+ wy0 10 16)))
    (setf (window-scroll-right-button-rect window) (make-rect :x0 (+ wx0 22 17) :y0 (+ wy0 10) :x1 (+ wx0 22 17 16) :y1 (+ wy0 10 16)))))

(defun draw-vertical-scroll-bar (window bar-percentage bar-start)
  (let* ((vert-rect (window-vertical-scroll-rect window))
	 (sx0 (rect-x0 vert-rect))
	 (sy0 (rect-y0 vert-rect))
	 (sy1 (rect-y1 vert-rect))
	 (ty0 (+ sy0 18 18))		;make room for scroll buttons
	 (ty1 (- sy1 1))
	 (vy (max 16 (* bar-percentage (- ty1 ty0))))
	 (vy0 (max ty0 (* (- 1.0 bar-start) (- ty1 vy)))))
    (if (= vy 16)
	(dimple-button-image sx0 vy0)
      (progn
	(dps-send (format nil ".667 setgray ~D ~D ~D ~D rectfill~%~
                               1 setgray ~D ~D moveto 0 ~D rlineto ~D 0 rlineto stroke~%"
			  (+ sx0 1) (+ vy0 1) 14 vy
			  (+ sx0 1) vy0 vy 14))
	(dimple-image (+ sx0 3) (+ vy0 (floor vy 2) -9))
	(dps-send (format nil ".333 setgray ~D ~D moveto ~D 0 rlineto 0 ~D rlineto stroke~%~
                               0 setgray    ~D ~D moveto ~D 0 rlineto 0 ~D rlineto stroke~%"
			  (+ sx0 2) (+ vy0 1) 13 (- vy 3)
			  (+ sx0 1) vy0 15 (- vy 1)))))))
	 
(defun draw-horizontal-scroll-bar (window bar-percentage bar-start)
  (let* ((horz-rect (window-horizontal-scroll-rect window))
	 (sx0 (rect-x0 horz-rect))
	 (sx1 (rect-x1 horz-rect))
	 (sy0 (rect-y0 horz-rect))
	 (tx0 (+ sx0 18 18))		;make room for scroll buttons
	 (tx1 (- sx1 1))
	 (vx (max 16 (* bar-percentage (- tx1 tx0))))
	 (vx0 (max tx0 (* (- 1.0 bar-start) (- tx1 vx)))))
    (if (= vx 16)
	(dimple-button-image vx0 sy0)
      (progn
	(dps-send (format nil ".667 setgray ~D ~D ~D ~D rectfill~%~
                               1 setgray ~D ~D moveto 0 ~D rlineto ~D 0 rlineto stroke~%"
			  (+ vx0 1) (+ sy0 1) vx 14
			  vx0 sy0 15 vx))
	(dimple-image (+ vx0 (floor vx 2) -6) (+ sy0 3))
	(dps-send (format nil ".333 setgray ~D ~D moveto ~D 0 rlineto 0 ~D rlineto stroke~%~
                               0 setgray    ~D ~D moveto ~D 0 rlineto 0 ~D rlineto stroke~%"
			  (+ vx0 2) (+ sy0 1) (- vx 3) 13
			  (+ vx0 1) sy0 (- vx 1) 14))))))

(defun draw-resize-bar (window)
  (let ((wx0 (window-x0 window))
	(wx1 (window-x1 window))
	(wy0 (window-y0 window)))
    (setf (window-resize-left-button-rect window) (make-rect :x0 (+ wx0 1) :y0 (+ wy0 1) :x1 (+ wx0 29) :y1 (+ wy0 6)))
    (setf (window-resize-right-button-rect window) (make-rect :x0 (- wx1 29) :y0 (+ wy0 1) :x1 (- wx1 1) :y1 (+ wy0 6)))
    (dps-send (format nil ".333 setgray ~D ~D ~D 1 rectfill~%~
                           1 setgray ~D ~D ~D 1 rectfill~%~
                           .667 setgray ~D ~D ~D ~D rectfill~%~
                           .333 setgray ~D ~D moveto 0 ~D rlineto stroke ~D ~D moveto 0 ~D rlineto stroke~%~
                           1 setgray ~D ~D moveto 0 ~D rlineto stroke ~D ~D moveto 0 ~D rlineto stroke~%~
                           0 setgray~%"
		      (+ wx0 1) (+ wy0 8) (- wx1 wx0 2)
		      (+ wx0 1) (+ wy0 7) (- wx1 wx0 2)
		      (+ wx0 1) (+ wy0 1) (- wx1 wx0 2) 6
		      (+ wx0 29) (+ wy0 2) 5 (- wx1 30) (+ wy0 2) 5
		      (+ wx0 30) (+ wy0 2) 5 (- wx1 29) (+ wy0 2) 5))))

(defun activate-window (window)
  ;; make the title-bar black (with white title)
  (draw-title-bar window t))

(defun deactivate-window (window)
  ;; make title-bar gray (with black title)
  (draw-title-bar window nil))

(defun scale-window (win new-scale)
  (setf (window-scale win) new-scale)
  (dps-send (format nil "grestore~%gsave [~D 0 0 ~D 0 0] concat~%" new-scale new-scale 0 0)))




(defun in-title-bar (window x y) (in-rect x y (window-title-bar-rect window)))
(defun in-close-button (window x y) (in-rect x y (window-close-button-rect window)))

(defvar mouse-up (dps-mouse-up))
(defvar mouse-down (dps-mouse-down))



;;; ----------- tracking rectangles

(defstruct rect x0 y0 x1 y1 obj owner matrix code data id depth)

(defun invert-2x2-matrix (mat)
  ;; needed for translation from window coordinates to object coordinates when scaling/transformation is taking place
  (let* ((a11 (first mat))
	 (a12 (second mat))
	 (a21 (third mat))
	 (a22 (fourth mat))
	 (ia11 (/ 1.0 a11))
	 (det (/ 1.0 (- a22 (* a21 a12 ia11)))))
    (list (+ ia11 (* ia11 a12 det a21 ia11))
	  (- (* det ia11 a12))
	  (- (* det ia11 a21))
	  det)))

(defun in-rect (x y rct)
  (and rct
       (<= (rect-y0 rct) y (rect-y1 rct))
       (<= (rect-x0 rct) x (rect-x1 rct))))

(defun rectless (r0 r1)
  (or (< (rect-x1 r0) (rect-x1 r1))
      (and (= (rect-x1 r0) (rect-x1 r1))
	   (or (< (rect-y1 r0) (rect-y1 r1))
	       (and (= (rect-y1 r0) (rect-y1 r1))
		    (or (> (rect-y0 r0) (rect-y0 r1))
			(and (= (rect-y0 r0) (rect-y0 r1))
			     (> (rect-x0 r0) (rect-x0 r1)))))))))

(defun sort-page (pg)
  (sort pg #'rectless))

(defun find-rectangle (x y rects &optional (depth 0) (id nil))
  (find-if #'(lambda (rct) 
	       (and (= depth (rect-depth rct)) 
		    (or (not id) 
			(not (member (rect-id rct) id)))
		    (in-rect x y rct)))
	   rects))

(defun find-clicked-rectangle (x y clicks max-depth rects)
  (let ((no-goods nil)
	(depth max-depth)
	(click (1- clicks))
	(last-fr nil))
    (loop do
      (let ((fr (find-rectangle x y rects depth no-goods)))
	(if fr
	    (if (zerop click)
		(return-from find-clicked-rectangle fr)
	      (progn
		(setf last-fr fr)
		(decf click)
		(push (rect-id fr) no-goods)))
	  (if (zerop depth)
	      (return-from find-clicked-rectangle last-fr)
	    (decf depth)))))))

(defun update-rectangle (x y r)
  (setf (rect-x0 r) (min x (or (rect-x0 r) x)))
  (setf (rect-y0 r) (min y (or (rect-y0 r) y)))
  (setf (rect-x1 r) (max x (or (rect-x1 r) x)))
  (setf (rect-y1 r) (max y (or (rect-y1 r) y))))

(defun edit-rectangle (rct x-incr y-incr)
  (incf (rect-x0 rct) x-incr)
  (incf (rect-y0 rct) y-incr)
  (incf (rect-x1 rct) x-incr)
  (incf (rect-y1 rct) y-incr))

(defun update-rectangles (server sx sy current-rectangles)	;turn image notion of coordinates into outer (window) view
  (let* ((new-edges (and (scaling-matrix server)
			 (transform-box (scaling-matrix server) 
					(list (fourth (scaling-matrix server)) (fifth (scaling-matrix server)))
					(list 0 0 sx sy))))
	 (x (or (and new-edges (third new-edges)) sx))
	 (y (or (and new-edges (fourth new-edges)) sy)))
    (loop for r in current-rectangles do (update-rectangle x y r))))

;;; after zoom or during scrolling, maybe only display when tracking-rectangle of object
;;; intersects current view -- visible-rect as object (unscaled) view of current visible coordinates

(defun rectangles-intersect (obj vis)
  (and (or (<= (rect-x0 vis) (rect-x0 obj) (rect-x1 vis))
	   (<= (rect-x0 vis) (rect-x1 obj) (rect-x1 vis))
	   (and (<= (rect-x0 obj) (rect-x0 vis)) 
		(<= (rect-x1 vis) (rect-x1 obj))))
       (or (<= (rect-y0 vis) (rect-y0 obj) (rect-y1 vis))
	   (<= (rect-y0 vis) (rect-y1 obj) (rect-y1 vis))
	   (and (<= (rect-y0 obj) (rect-y0 vis))
		(<= (rect-y1 vis) (rect-y1 obj))))))

(defun copy-rectangle (rct)
  (make-rect :x0 (rect-x0 rct) :y0 (rect-y0 rct) :x1 (rect-x1 rct) :y1 (rect-y1 rct)))


(defun create-window (x0 y0 x1 y1 &key (title "Untitled") menu event-mask
				     (close-button t) (vertical-scroll nil) (horizontal-scroll nil) (resize nil))
  (let* ((win (make-window :title title :active t :menu-rects menu))
	 (wx0 0)
	 (wy0 0)
	 (wx1 (- (min x1 screen-width) x0))	;DPS window args are lower-left corner width height
	 (wy1 (- (min y1 screen-height) y0)))

    (setf (window-x0 win) wx0)
    (setf (window-x1 win) wx1)
    (setf (window-y0 win) wy0)
    (setf (window-y1 win) wy1)

    (when menu
      (let ((mr (copy-rectangle (first menu))))
	(loop for rct in (cdr menu) by #'cdr do
	  (setf (rect-x0 mr) (min (rect-x0 mr) (rect-x0 rct)))
	  (setf (rect-y0 mr) (min (rect-y0 mr) (rect-y0 rct)))
	  (setf (rect-x1 mr) (max (rect-x1 mr) (rect-x1 rct)))
	  (setf (rect-y1 mr) (max (rect-y1 mr) (rect-y1 rct))))
	(setf (window-menu-rect win) mr)))

    ;;get window server connection
    (dps-create-context)

    ;; set up the window, display it on top
    (dps-send (format nil "/showpage {initgraphics} bind def~%~
                           ~D ~D ~D ~D 0 window~%~
                           windowdeviceround~%~
                           1 0 currentwindow orderwindow~%"
		      x0 y0 wx1 wy1))

    (dps-send (format nil "~D currentwindow seteventmask~%"
		      (or event-mask
			  (logior (dps-mouse-down-mask) (dps-mouse-up-mask)))))

    (dps-send (format nil "0 1 ~D ~D rectstroke~%" (- wx1 1) (- wy1 1)))
    (activate-window win)
    (setf (window-title-bar-rect win) (make-rect :x0 wx0 :y0 (- wy1 23) :x1 (- wx1 24) :y1 wy1))
    (if close-button
	(setf (window-close-button-rect win) (make-rect :x0 (- wx1 23) :y0 (- wy1 23) :x1 (- wx1 1) :y1 (- wy1 1))))

    (setf (window-rx0 win) (+ wx0 1 (if vertical-scroll 19 0)))
    (setf (window-ry0 win) (+ wy0 1 (if resize 8 0) (if horizontal-scroll 20 0)))
    (setf (window-rx1 win) (- wx1 2))
    (setf (window-ry1 win) (- wy1 22 1)) ;was also 20 18 (menu?)
    (setf (window-user-rect win) 
      (make-rect :x0 (window-rx0 win) :y0 (window-ry0 win) :x1 (window-rx1 win) :y1 (window-ry1 win)))

    (if vertical-scroll (draw-vertical-scroll-bar-slot win))
    (if resize (draw-resize-bar win))
    (if horizontal-scroll (draw-horizontal-scroll-bar-slot win))

    (if menu
	(loop for rct in menu do (draw-menu-item (rect-data rct) rct nil)))

    win))



(defun draw-menu-item (title mrect inverse)
  ;; mrect is the menu-item-rectangle
  (let ((x0 (rect-x0 mrect))
	(y0 (rect-y0 mrect))
	(x1 (rect-x1 mrect))
	(y1 (rect-y1 mrect)))
    (setf (rect-owner mrect) title)
    (dps-send (format nil "~1,3F setgray ~D ~D ~D ~D rectfill~%~
                           1 setgray ~D ~D moveto 0 ~D rlineto ~D 0 rlineto stroke~%~
                           0 setgray ~D ~D moveto ~D 0 rlineto 0 ~D rlineto stroke gsave~%~
                           ~D ~D moveto ~D setgray~%~
                           /~A findfont ~D scalefont setfont~%~
                           (~A) show grestore~%~
                           0 setgray~%"
		      (if inverse 1.0 .667) 
		      x0 y0 (- x1 x0) (- y1 y0)
		      x0 y0 (- y1 y0) (- x1 x0)
		      x0 y0 (- x1 x0) (- y1 y0)
		      (+ x0 5) (+ y0 4)
		      (if inverse .333 0)
		      "Helvetica-Narrow" (- y1 y0 6) title))
    (if inverse (dps-flush))))

(defun perform-menu-action (win x y)
  (let ((rct (find-if #'(lambda (r) (in-rect x y r)) (window-menu-rects win))))
    (when (and rct (rect-code rct))
      (unprotect-window win)
      (draw-menu-item (rect-owner rct) rct t)
      (protect-window win)
      (funcall (rect-code rct))
      (unprotect-window win)
      (draw-menu-item (rect-owner rct) rct nil)
      (protect-window win))))




(defun run-window (win user-command-handler lisp-handler exit-actions)
  (protect-window win)
  (setf (window-unsaved-changes win) nil)
  (let ((happy t)
	(success 0))
    (unwind-protect
	(loop while happy do
	  (dps-flush)
	  (setf success (dps-get-event .5))
	  ;; no recent mouse events -- look for typing in lisp window and deal with it.
	  ;; major danger here is an attempt to call cmn while we're editing the cmn structure:
	  ;; not only will we get very confused, but we can easily kill the window server.
	  (if (zerop success)
	      (if (listen)
		  (let* ((val (eval (read-from-string (read-line)))))
		    (if lisp-handler 
			(print (funcall lisp-handler val))
		      (print val))
		    (terpri)
		    (force-output)))
	    (let ((type (dps-event-type))
		  (mx (dps-event-location-x))
		  (my (dps-event-location-y))
		  (clicks (dps-event-mouse-click)))

	      (if (and (= type mouse-up)
		       (plusp clicks))
		  (if (in-close-button win mx my)
		      ;; click in the close-button closes window
		      (setf happy nil)
		    (if (in-rect mx my (window-menu-rect win))
			(perform-menu-action win mx my)))
		      
		(if (= type mouse-down)
		    (progn
		      (if (not (window-active win)) (activate-window win))
		      ;; make sure we're on top too
		      (dps-send (format nil "1 0 ~D orderwindow~%" (dps-event-window)))
		      (if (and user-command-handler
			       (in-rect mx my (window-user-rect win)))
			  (funcall user-command-handler win mx my clicks)
			(if (in-title-bar win mx my)
			    ;; drag in title-bar moves entire window
			    ;; I don't know what "workspaceWindow" is, or why this code works (there must be some
			    ;;   translation from relative mouse coordinates to movewindow coordinates).  No mention of
			    ;;   this window in /usr/lib/NextStep/windowPackage.ps.
			    (dps-send (format nil "{~D stilldown not {exit} if~%~
				                   workspaceWindow currentmouse ~D sub exch ~D sub exch~%~
				                   ~D movewindow} loop~%"
					      (dps-event-mouse-eventnum)
					      my mx
					      (dps-event-window)))))))))))
      
      (progn				;unwind-protect exit block
	(if exit-actions (funcall exit-actions win))
	(dps-send (format nil "grestore~%"))
	(dps-send (format nil "~D termwindow~%nulldevice~%" (dps-event-window)))
	(dps-flush)
	(dps-kill-context)))))



;;; an example that just sets up a window and lets us push it around on the screen

(defun display-file (file x0 y0 x1 y1 &optional initial-actions commands exit-actions)
  (let ((win (create-window x0 y0 x1 y1)))
    (protect-window win)
    (if initial-actions (funcall initial-actions))
    (dps-display-file file)
    (run-window win commands nil exit-actions)))

