;;; -*- Lisp -*-

(in-package :cmn)

;;; this file provides the wedged beams used by folks like bartok
;;; there are two versions, a simple beam addition, and a whole new cmn object type


;;; the simplest version just adds the slanted beams to a normal beam:

(defun display-wedge-beams (mark note score start-beams end-beams &rest rest)
  (declare (ignore mark))
  (when (and (not (member :just-fooling rest))
	     (not (member :just-do-it rest)))
    ;; now find the beam location and draw the other beams
    (let* ((beamxy (outer-beam note))
	   (beam-above (eq (beam-direction (beams note)) :up))
	   (xb (+ (first beamxy) (half-stem-width score)))
	   (xe (third beamxy))
	   (yb (second beamxy))
	   (ye (fourth beamxy))
	   (faster (< start-beams end-beams))
	   (beam-sep1 (beam-spacing score))
	   (beam-wid1 (beam-width score))
	   (beam-sep (if beam-above (- beam-sep1) beam-sep1))
	   (beam-thickness (if beam-above beam-wid1 (- beam-wid1)))
	   (dyb (if faster 0 beam-sep))
	   (dye (if faster beam-sep 0))
	   (beams (abs (- start-beams end-beams))))
      (when (not beam-above)
	(progn
	  (incf yb beam-wid1)
	  (incf ye beam-wid1)))
      (if (> (min start-beams end-beams) 1)
	  (let ((base-beam-sep (* (1- (min start-beams end-beams)) beam-sep)))
	    (incf ye base-beam-sep)
	    (incf yb base-beam-sep)))
      (loop for i from 1 to beams and y0 from (+ yb dyb) by dyb and y1 from (+ ye dye) by dye do
	;; now draw the beam with a sharp point at the end with fewer beams
	(moveto score xb y0)
	(fill-in score
		 :path (list xb y0 
			     xe y1
			     xe (+ y1 (if faster beam-thickness 0))
			     xb (+ y0 (if faster 0 beam-thickness))))))))

(defun begin-wedge-beam (start-beams end-beams &rest args)
  (engorge
   (list
    (apply #'begin-beam no-beam-break
	                (dy (list (* (+ .5 (max start-beams (1- end-beams))) 1.25 .126)
				  (* (+ .5 (max end-beams (1- start-beams))) 1.25 .126)))
	                args)
    (mark #'(lambda (mark note score &rest rest)
	      (apply #'display-wedge-beams mark note score start-beams end-beams rest))
	  :wedge-beam))))

(defun end-wedge-beam (&optional ur-beam)
  (end-beam ur-beam))

#|
(cmn staff treble (c4 e (begin-wedge-beam 4 1)) (c4 e) (c4 e) (c4 e (end-wedge-beam)))
(cmn staff treble (c5 e (begin-wedge-beam 4 1)) (c5 e) (c5 e) (c5 e (end-wedge-beam)))
(cmn staff treble (c4 e (begin-wedge-beam 1 4)) (c4 e) (c4 e) (c4 e (end-wedge-beam)))
(cmn staff treble (c4 e (begin-wedge-beam 1 4)) (c4 e) (d4 e) (e4 e) (f4 e) (d4 e) (e4 e) (g4 e (end-wedge-beam)))
|#



;;; but that version doesn't know what the duration of the overall group is (for metering purposes and
;;; so on), nor does it automatically provide the changed spacing of the notes within the group), so
;;; the following version treats the group under the beam as a sort of chord spread out horizontally.
;;; To keep other simultaneous voices lined up right, we insert invisible spacers into the current staff
;;; and fix up their apparent x0 and x1 later.


(defvar spacer-space 0)      ;a bit of a kludge (else need pointers into current staff data)

(defclass spacer (wrest)
  ())

(defmethod house ((sp spacer) score)
  (declare (ignore score))
  (setf (box-x1 sp) spacer-space))

(defmethod display ((sp spacer) container score &rest rest)
  (declare (ignore container score rest))
  nil)

(defun invisible-spacer (dur)
  (make-instance 'spacer :quarters dur))



(defclass bartok-bounce (chord)
  ((data :initarg :data :initform nil :accessor bartok-bounce-data)
   (spacing :initarg :spacing :initform .7 :accessor spacing)
   (start-beams :initarg :start-beams :initform 1 :accessor start-beams)
   (end-beams :initarg :end-beams :initform 1 :accessor end-beams)))

(self-action start-beams setf-start-beams)
(self-action end-beams setf-end-beams)
(self-action spacing setf-spacing)

(defmethod bartok-bounce-p ((obj t)) nil)
(defmethod bartok-bounce-p ((obj bartok-bounce)) t)

(defun bartok-bounce (&rest objects)
  (let ((notes nil)
	(bb (make-instance 'bartok-bounce)))
    (setf *cmn-owning-object* bb)
    (loop for object in objects do
      (setf *cmn-object* object)
      (when object
	(if (self-acting-p object)
	    (funcall (action object) bb (arguments object))
	  (if (note-p object)
	      (push object notes)
	    (if (rhythm-p object)
		(rcopy object bb)
	      (if (score-object-p object)
		  (if (or (sundry-p object) (pause-p object) (dynamics-p object) (glyph-list-p object))
		      (add-to-marks bb (list (if (write-protected object) (copy object) object)))
		    (copy object bb))
		(if (score-object-list-p object)
		    (loop for note in (data object) do
		      (push (if (write-protected note) (copy note) note) notes))
		  (if (text-p object)
		      (add-to-marks bb (list object))
		    (cmn-warn "odd argument to bartok-bounce: ~A" object)))))))))
    (setf (bartok-bounce-data bb) (nreverse notes))
    (loop for note in (bartok-bounce-data bb) do
      (setf (beat note) 0)
      (setf (flags note) 1)
      (setf (quarters note) 1/2))
    (cmn-tick-pipe bb)
    (setf (quarters bb) (/ (quarters bb) 4))
    (engorge (list
	      bb 
	      (invisible-spacer (quarters bb))
	      (invisible-spacer (quarters bb))
	      (invisible-spacer (quarters bb))))))

(defmethod display ((bb bartok-bounce) container score &rest rest)
  (declare (ignore container))
  ;; set position of each note, beam the bb, etc
  ;; if this is a housing pass, just figure out how much room we'll need at a minimum
  (let* ((notes (length (bartok-bounce-data bb)))
	 (spaces (1- notes))
	 (note-linear-dxs (loop for note in (bartok-bounce-data bb)
			   collect (if (note-p note)
				       (+ .5 
					  (if (note-sign note) (+ (g-rx (note-sign note)) (accidental-breathing-space score)) 0)
					  (if (or (< (line note) -1) 
						  (> (line note) 9))
					      .1
					    0))
				     (+ .5 
					(if (find-if #'sign (chord-data note))
					    .3
					  0)))))
	 (linear-x-space (+ .1 (apply #'+ note-linear-dxs)))
	 (faster (> (end-beams bb) (start-beams bb)))
	 (note-dxs (append (list 0)
			    (let* ((space (* .5 linear-x-space))
				   (spacex (loop for n from 1 to spaces 
					    collect (* space (spacing bb))
					    do (decf space (* space (spacing bb))))))
			      (if (not faster) (setf spacex (nreverse spacex)))
			      (loop for x in note-linear-dxs and dx in spacex 
			       collect (+ x dx)))))
	 (x-space (apply #'+ note-dxs)))
    (if (member :just-fooling rest)
	(moveto score (+ (box-x0 bb) (setf spacer-space (/ x-space 4))) (box-y0 bb))
      (let ((x0 (box-x0 bb))
	    (stfy0 (%staff-y0 bb)))
	(loop for note in (bartok-bounce-data bb) and dx in note-dxs do
	  (setf (box-x0 note) (+ x0 dx))
	  (incf x0 dx)
	  (setf (box-y0 note) (* (minimum-line note) (staff-line-separation score)))
	  (setf (center note) (+ .15 
				 (if (and (note-sign note) (zerop (audible-collision note)))
				     (+ (g-rx (note-sign note))
					(if (or (<= (minimum-line note) -2) (<= 10 (maximum-line note)))
					    (leger-line-length score)
					  0)
					(accidental-breathing-space score))
				   0)))
	  (setf (%staff-y0 note) stfy0))
	(annotate-beam score nil stfy0 (first (bartok-bounce-data bb)) (bartok-bounce-data bb) 1.0)
	(setf (dxy-dy (beams (first (bartok-bounce-data bb)))) 
	  (list (* (+ .5 (max (start-beams bb) (1- (end-beams bb)))) 1.25 .126)
		(* (+ .5 (max (end-beams bb) (1- (start-beams bb)))) 1.25 .126)))
	(add-to-marks (first (bartok-bounce-data bb))
		      (list (mark #'(lambda (mark note score &rest rest)
				      (apply #'display-wedge-beams mark note score 
					     (start-beams bb) 
					     (end-beams bb)
					     rest))
				  :wedge-beam)))
	(loop for note in (bartok-bounce-data bb) do
	  (display note nil score))))))

;;; now set up the wedged beam stuff as before
;;; (cmn staff treble (bartok-bounce q (notes c4 d4 e4 f4) (start-beams 1) (end-beams 3)))
#|
 (cmn (staff treble (c4 q) (bartok-bounce q (notes c4 d4 e4 f4 e4 d4 c4) (spacing .7) (start-beams 1) (end-beams 3)) (c4 q)) 
      (staff bass c3 e c3 e c3 e c3 e c3 e c3 e))

;;; the arguments to bartok-bounce are the duration (and onset time if desired)
;;;    the list of notes under the wedged beams (in the "notes" statement) -- each of these is a full fledged note
;;;    start-beams and end-beams (as above)
;;;    spacing = how to space the notes -- between 0 and 1, default is .7 -- clumpier as you approach 1

 (cmn (staff treble (meter 2 4) 
        (c4 q) 
        (bartok-bounce q (notes (c4 staccato) (d4 marcato) 
                                (e4 upside-down-fermata) (f4 (bartok-pizzicato (dy .5))) 
                                (e4 (fingering 7 9)) (d4 ppp) c4) 
          (spacing .5) (start-beams 1) (end-beams 3)) 
        (c4 q)) 
      (staff bass (meter 2 4) c3 e c3 e c3 e c3 e c3 e c3 e))


(cmn (size 24) (AUTOMATIC-LINE-BREAKS nil) 
  (staff treble       
    (note fs5 s (begin-wedge-beam 2 4))(note fn5 s)(note e5 s)(note fs5 s) 
    (note fn5 s) (note e5 s) (note fs5 s) (note fn5 s)
    (note e5 s) (note fs5 s) (note fn5 s) (note e5 s)
    (note fs5 s) (note fn5 s) (note e5 s) (note fs5 s)
    (note fn5 s) (note e5 s) (note fs5 s)(note fn5 s)
    (note e5 s) (note fs5 s) (note fn5 s) (note e5 s (end-wedge-beam))
    (bar)(line-mark)))

|#